@@ -1,4 +1,4 @@
-{-# Language OverloadedStrings #-}
+{-# Language OverloadedStrings, BlockArguments #-}
 {-|
 Module      : Irc.Commands
 Description : Smart constructors for "RawIrcMsg"
@@ -18,6 +18,7 @@ module Irc.Commands
   , ircCnotice
   , ircCprivmsg
   , ircEtrace
+  , ircHelp
   , ircInfo
   , ircInvite
   , ircIson
@@ -34,9 +35,11 @@ module Irc.Commands
   , ircMode
   , ircMonitor
   , ircMotd
+  , ircNames
   , ircNick
   , ircNotice
   , ircOper
+  , ircOperwall
   , ircPart
   , ircPass
   , ircPing
@@ -51,10 +54,16 @@ module Irc.Commands
   , ircTime
   , ircTopic
   , ircTrace
+  , ircUndline
   , ircUnkline
+  , ircUnxline
+  , ircUnresv
   , ircUser
   , ircUserhost
+  , ircUserip
+  , ircUsers
   , ircVersion
+  , ircWallops
   , ircWho
   , ircWhois
   , ircWhowas
@@ -80,7 +89,7 @@ import qualified Data.ByteString as B
 import qualified Data.ByteString.Base64 as Enc
 
 nonempty :: Text -> [Text]
-nonempty txt = filter (not . Text.null) [txt]
+nonempty txt = [txt | not (Text.null txt)]
 
 -- | PRIVMSG command
 ircPrivmsg ::
@@ -161,6 +170,12 @@ ircWallops ::
   RawIrcMsg
 ircWallops msg = rawIrcMsg "WALLOPS" [msg]
 
+-- | OPERWALL command
+ircOperwall ::
+  Text {- ^ message -} ->
+  RawIrcMsg
+ircOperwall msg = rawIrcMsg "OPERWALL" [msg]
+
 -- | NICK command
 ircNick ::
   Text {- ^ nickname -} ->
@@ -222,8 +237,42 @@ ircKline minutes mask reason = rawIrcMsg "KLINE" [minutes, mask, reason]
 -- | UNKLINE command
 ircUnkline ::
   Text {- ^ mask -} ->
+  Maybe Text {- ^ server -} ->
   RawIrcMsg
-ircUnkline mask = rawIrcMsg "UNKLINE" [mask]
+ircUnkline mask mbServer = rawIrcMsg "UNKLINE"
+  case mbServer of
+    Nothing     -> [mask]
+    Just server -> [mask, "ON", server]
+
+-- | UNDLINE command
+ircUndline ::
+  Text {- ^ mask -} ->
+  Maybe Text {- ^ server -} ->
+  RawIrcMsg
+ircUndline mask mbServer = rawIrcMsg "UNDLINE"
+  case mbServer of
+    Nothing     -> [mask]
+    Just server -> [mask, "ON", server]
+
+-- | UNXLINE command
+ircUnxline ::
+  Text       {- ^ mask -} ->
+  Maybe Text {- ^ server -} ->
+  RawIrcMsg
+ircUnxline mask mbServer = rawIrcMsg "UNXLINE"
+  case mbServer of
+    Nothing     -> [mask]
+    Just server -> [mask, "ON", server]
+
+-- | UNRESV command
+ircUnresv ::
+  Text       {- ^ mask   -} ->
+  Maybe Text {- ^ server -} ->
+  RawIrcMsg
+ircUnresv mask mbServer = rawIrcMsg "UNRESV"
+  case mbServer of
+    Nothing     -> [mask]
+    Just server -> [mask, "ON", server]
 
 -- | TESTLINE command
 ircTestline ::
@@ -255,7 +304,7 @@ ircChantrace channel = rawIrcMsg "CHANTRACE" [channel]
 ircEtrace ::
   Text {- ^ argument -} ->
   RawIrcMsg
-ircEtrace arg = rawIrcMsg "ETRACE" [arg]
+ircEtrace arg = rawIrcMsg "ETRACE" (nonempty arg)
 
 -- | REMOVE command
 ircRemove ::
@@ -267,7 +316,7 @@ ircRemove chan who msg = rawIrcMsg "REMOVE" (idText chan : who : nonempty msg)
 
 -- | QUIT command
 ircQuit :: Text {- ^ quit message -} -> RawIrcMsg
-ircQuit = rawIrcMsg "QUIT" . nonempty
+ircQuit msg = rawIrcMsg "QUIT" [msg]
 
 -- | PASS command
 ircPass :: Text {- ^ password -} -> RawIrcMsg
@@ -404,6 +453,18 @@ ircUser ::
   Text {- ^ realname -} -> RawIrcMsg
 ircUser user real = rawIrcMsg "USER" [user, "0", "*", real]
 
+-- | NAMES command
+ircNames ::
+  Identifier {- ^ channel -} ->
+  RawIrcMsg
+ircNames chan = rawIrcMsg "NAMES" [idText chan]
+
+-- | HELP command
+ircHelp ::
+  Text {- ^ topic -} ->
+  RawIrcMsg
+ircHelp topic = rawIrcMsg "HELP" [topic]
+
 -- | CAP REQ command
 ircCapReq ::
   [Text] {- ^ capabilities -} ->
